/*
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Copyright 2000 by Chris Hertel, Michael B. Allen
 *
 */

package jcifs.netbios;

class Name {
    private final static char[] HEX_DIGITS = { 
    '0' , '1' , '2' , '3' , '4' , '5' ,
    '6' , '7' , '8' , '9' , 'A' , 'B' ,
    'C' , 'D' , 'E' , 'F'
    };

    byte[] wireFormat;
    String name;
    int type;
    String scope;
    int wireLength = 34;
    int hash;

    Name() {
        wireFormat = new byte[wireLength];
    }
    Name( String name, int type, String scope ) {
        wireFormat = new byte[wireLength];
        setName( name );
        setType( type );
        setScope( scope );
    }

    void setName( String n ) {

        if( n == null ) {
            name = null;
            return;
        }
        name = ( n.length() > 15 ) ? n.substring( 0, 15 ).toUpperCase() : n.toUpperCase();

        // write 0x20 in first byte
        wireFormat[0] = 0x20;

        // write name
        byte tmp[] = name.getBytes();
        int i;
        for( i = 0; i < tmp.length; i++ ) {
            wireFormat[2*i+1] = (byte)((( tmp[i] & 0xF0 ) >> 4 ) + 0x41 );
            wireFormat[2*i+2] = (byte)(( tmp[i] & 0x0F ) + 0x41 );
        }
        for( ; i < 15; i++ ) {
            wireFormat[2*i+1] = 0x43;
            wireFormat[2*i+2] = 0x41;
        }
    }
    void setType( int type ) {

        this.type = type;

        // write type
        wireFormat[31] = (byte)((( type & 0xF0 ) >> 4 ) + 0x41 );
        wireFormat[32] = (byte)(( type & 0x0F ) + 0x41 );
    }
    void setScope( String s ) {
        int newScopeLength;

        if( scope != null ) {
            if( s == null || ( newScopeLength = s.length() ) == 0 ) {
                // scope wasn't null but now it is so let's trunc
                wireLength = 34;
                byte tmp[] = wireFormat;
                wireFormat = new byte[wireLength];
                System.arraycopy( tmp, 0, wireFormat, 0, 33 );
                wireFormat[33] = (byte)0x00;
                scope = null;
                return;
            }
        } else if( s == null || ( newScopeLength = s.length() ) == 0 ) {
            // both scope's are null so nothing to do
            return;
        }

        // if the scopes are not the same size, adjust the wireFormat array
        if( scope == null || newScopeLength != scope.length() ) {
            wireLength = 35 + newScopeLength;
            byte tmp[] = wireFormat;
            wireFormat = new byte[wireLength];
            System.arraycopy( tmp, 0, wireFormat, 0, 33 );
        }

        // copy new scope in
        wireFormat[33] = (byte)'.';
        System.arraycopy( s.getBytes(), 0, wireFormat, 34, newScopeLength );

        // now go over scope backwards converting '.' to label length
        int c = 0;
        for( int j = wireLength - 2; j > 32; j-- ) {
            if( wireFormat[j] == '.' ) {
                wireFormat[j] = (byte)c;
                c = 0;
            } else {
                c++;
            }
        }
        // write trailing 0x00
        wireFormat[wireLength - 1] = (byte)0x00;

        // save new scope member variable
        scope = s;
    }
    public String toString() {
        StringBuffer sb = new StringBuffer();
        // fix MSBROWSE name
        if( name.charAt( 0 ) == 0x01 ) {
            byte b[] = name.getBytes();
            b[0] = '.';
            b[1] = '.';
            b[14] = '.';
            sb.append( new String( b ));
        } else {
            sb.append( name );
        }
        sb.append( "<" ).append( toHexChars( type )).append( ">" );
        if( scope != null && scope.length() != 0 ) {
            sb.append( "." ).append( scope );
        }
        return sb.toString();
    }
    public int hashCode() {
        hash = 0;

        // Compute name and type portion of hash code
        int length = 33;
        for( int i = 0; i < length; i++ ) {
            hash = wireFormat[i] + 65599 * hash;
        }

        if( wireLength > 34 ) {
            // Factor scope into hash code
            for( int j = 33; j < wireLength; j++ ) {
                hash = wireFormat[j] + 65599 * hash;
            }
        }
        return hash;
    }
    public boolean equals( Object obj ) {
        if( !( obj instanceof Name )) {
            return false;
        }
        Name n = (Name)obj;
        // Make a few simple checks
        if( type != n.type || n.wireLength != wireLength ) {
            return false;
        }
        // Check the name
        for( int i = 1; i < 31; i++ ) {
            if( n.wireFormat[i] != wireFormat[i] ) {
                return false;
            }
        }
        // Comes down to scope
        for( int i = 34; i < wireLength; i++ ) {
            if( n.wireFormat[i] != wireFormat[i] ) {
                return false;
            }
        }
        return true;
    }

    // Populate class variables name, type and
    // scope from wire format array

    int readFullName( byte[] src, int srcIndex ) {

        // find wire length
        if(( src[srcIndex + 33] & 0x3F ) != 0 ) {
            wireLength = wireLength( src, srcIndex );
            wireFormat = new byte[wireLength];
        }
        System.arraycopy( src, srcIndex, wireFormat, 0, wireLength );

        byte tmp[] = new byte[wireLength];

        // read name
        int length = 15;
        for( int i = 0; i < 15; i++ ) {
            tmp[i] = (byte)((( wireFormat[2*i+1] & 0xFF ) - 0x41 ) << 4 );
            tmp[i] |= (byte)((( wireFormat[2*i+2] & 0xFF ) - 0x41 ) & 0x0F );
            if( tmp[i] != ' ' ) {
                length = i + 1;
            }
        }
        name = new String( tmp, 0, length );

        // read type
        type = (( wireFormat[31] & 0xFF ) - 0x41 ) << 4;
        type |= (( wireFormat[32] & 0xFF ) - 0x41 ) & 0x0F;

        // read scope
        if( wireLength > 34 ) {
            System.arraycopy( wireFormat, 33, tmp, 0, wireLength - 33 );
            int i = 0;
            while(( length = tmp[i] & 0x3F ) != 0 ) {
                tmp[i] = (byte)'.';
                i += length + 1;
            }
            scope = new String( tmp, 1, wireLength - 35 );
        } else {
            scope = new String();
        }
        return wireLength;
    }
    int writeFullName( byte[] dst, int dstIndex ) {
        System.arraycopy( wireFormat, 0, dst, dstIndex, wireLength );
        return wireLength;
    }
    int writeScope( byte[] dst, int dstIndex ) {
        System.arraycopy( wireFormat, 33, dst, dstIndex, wireLength - 33 );
        return wireLength - 33;
    }
    static int wireLength( byte[] src, int srcIndex ) {
        if(( src[srcIndex + 33] & 0x3F ) == 0 ) {
            return 34;
        }
        // then it has scope so walk down labels to find length
        int l, i = 33;
        while(( l = src[srcIndex + i] & 0x3F ) != 0 ) {
            i += l + 1;
        }
        return i + 1;
    }
    static String toHexChars( int val ) {
        int len = 2;
        byte[] tmp = new byte[2];
        while( len > 0 ) {
            tmp[ len - 1 ] = (byte)HEX_DIGITS[ val & 0x000F ];
            if( val != 0 ) {
                val >>>= 4;
            }
            len--;
        }
        return new String( tmp );
    }
}

